import os
import itab
import logging

from intervaltree import IntervalTree
from tqdm import tqdm
from collections import defaultdict

from bgparsers.readers.common import __count_lines

ELEMENTS_HEADER = ['CHROMOSOME', 'START', 'STOP', 'STRAND', 'ELEMENT', 'SEGMENT', 'SYMBOL']
"""
Headers of the data expected in the elements file (see :class:`~oncodrivefml.main.OncodriveFML`).
"""

ELEMENTS_SCHEMA = {
    'fields': {
        'CHROMOSOME': {'reader': 'str(x)', 'validator': "x in ([str(c) for c in range(1,23)] + ['X', 'Y', 'M'])"},
        'START': {'reader': 'int(x)', 'validator': 'x > 0'},
        'STOP': {'reader': 'int(x)', 'validator': 'x > 0'},
        'STRAND': {'reader': 'str(x)', 'validator': "x in ['.', '+', '-']"},
        'ELEMENT': {'reader': 'str(x)'},
        'SEGMENT': {'reader': 'str(x)', 'nullable': 'True'},
        'SYMBOL': {'reader': 'str(x)', 'nullable': 'True'}
}}


def elements(file, show_progress=True):
    """
    Parse an elements file compliant with :attr:`ELEMENTS_HEADER`

    Args:
        file: elements file. If 'SEGMENT' field is not present, the value of the 'ELEMENT'
            is used instead.

    Returns:
        dict: elements' (see :ref:`elements <elements dict>`).

    """

    lines_count = __count_lines(file) if show_progress else None

    elements = {}
    with itab.DictReader(file, header=ELEMENTS_HEADER, schema=ELEMENTS_SCHEMA) as reader:
        all_errors = []
        msg = "'{}'".format(os.path.basename(file)) if type(file) == str else ""
        for i, (r, errors) in enumerate(tqdm(reader, total=lines_count, desc="Parsing elements {}".format(msg).rjust(40), disable=(not show_progress)), start=1):
            # Report errors and continue
            if len(errors) > 0:
                all_errors += errors
                continue

            r['LINE'] = i

            elements[r['ELEMENT']] = elements.get(r['ELEMENT'], []) + [r]

        if len(all_errors) > 0:
            logging.warning("There are {} errors at {}. {}".format(
                len(all_errors), os.path.basename(file),
                " I show you only the ten first errors." if len(all_errors) > 10 else ""
            ))
            for e in all_errors[:10]:
                logging.warning(e)

    return elements


def elements_tree(file):
    e = elements(file)
    regions_tree = defaultdict(IntervalTree)
    for i, (k, allr) in enumerate(tqdm(e.items(), total=len(e), desc="Building mapping tree".rjust(40))):
        for r in allr:
            regions_tree[r['CHROMOSOME']][r['START']:(r['STOP']+1)] = (r['ELEMENT'], r['LINE'])
    return regions_tree




import functools
import os
import sys
import logging
from multiprocessing.pool import ThreadPool

import click

from multiprocessing import Pool
from os.path import isdir, join, basename, dirname, exists
from tqdm import tqdm

from bgparsers import selector
from bgparsers.commands import tasks
from bgparsers.utils import dequote


logger = logging.getLogger("bgparsers")


def __preprocess_chunk(file, chunk):
    return join(dirname(file), "bgvariants", "preprocess", basename(file), "c{:06d}.bgvars.xz".format(chunk))


def __input_to_selection(input):

    if len(input) == 0:
        input = (os.getcwd(), )

    selection = []
    for i in input:
        selection += (list(selector.find(i)) if isdir(i) else [(i, None)])

    # TODO Implement this at selector
    selection_chunks = []
    for file, ann in selection:
        chunk = 0
        chunk_file = __preprocess_chunk(file, chunk)
        if exists(chunk_file):
            while exists(chunk_file):
                selection_chunks.append((chunk_file, ann))
                chunk += 1
                chunk_file = __preprocess_chunk(file, chunk)
        else:
            selection_chunks.append((file, ann))

    return selection_chunks


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.option('--debug', help='Show debug messages.', is_flag=True)
@click.version_option()
def cli(debug):

    # Configure the logging
    logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s', datefmt='%H:%M:%S')
    logging.getLogger("bgparsers").setLevel(logging.DEBUG if debug else logging.INFO)
    logger.debug(sys.argv)


@click.command(short_help='Create a preprocessed version of the input file/folder.')
@click.argument('input', type=click.Path(exists=True), nargs=-1)
@click.option('--force', help='Overwrite files if exists.', is_flag=True)
@click.option('--cores', help='Maximum processes to run in parallel.', type=click.INT, default=os.cpu_count())
def preprocess(input, force, cores):

    selection = __input_to_selection(input)
    pool_method = Pool if len(selection) > 1 and cores > 1 else ThreadPool

    with pool_method(cores) as pool:
        task = functools.partial(tasks.preprocess, force=force, exist_ok=True)
        for output_file in tqdm(pool.imap_unordered(task, selection), total=len(selection),  desc="Parsing input files".rjust(40)):
            logger.debug(output_file)


def __parse_where(where):
    where_dict = {}
    for w in where:
        if "==" in w:
            field, value = w.split("==")
            where_dict[field.strip()] = dequote(value.strip())
            continue

        logger.error("Unknown where syntax '%s'", w)
        sys.exit(-1)
    return where_dict


@click.command(short_help='Concatenate files to standard input')
@click.argument('input', type=click.Path(exists=True), nargs=-1)
@click.option('--columns', '-c', multiple=True, type=click.STRING, help="Extra columns to add")
@click.option('--where', '-w', multiple=True, type=click.STRING, help="Filter expression. ie: CHROMOSOME==\"4\"")
def cat(input, columns, where):
    tasks.cat(
        __input_to_selection(input),
        columns=columns,
        where=__parse_where(where)
    )


@click.command(short_help='Count number of variants')
@click.argument('input', type=click.Path(exists=True), nargs=-1)
@click.option('--where', '-w', multiple=True, type=click.STRING)
@click.option('--groupby', '-g', type=click.STRING)
@click.option('--cores', help='Maximum processes to run in parallel.', type=click.INT, default=os.cpu_count())
@click.option('--quite', '-q', help="Don't show the progress, only the total count.", is_flag=True)
def count(input, where, groupby, cores, quite):

    selection = __input_to_selection(input)
    with Pool(cores) as pool:
        task = functools.partial(tasks.count, where=__parse_where(where), groupby=groupby)
        map_method = pool.imap_unordered if len(selection) > 1 else map

        total = 0
        groups = {}
        for c, g in tqdm(
                map_method(task, selection),
                total=len(selection),
                desc="Counting variants".rjust(40),
                disable=(len(selection) < 2 or quite)
        ):
            # Update groups
            if g is not None:
                for k, v in g.items():
                    val = groups.get(k, 0)
                    groups[k] = val + v

            total += c

    if len(groups) > 0:
        for k, v in sorted(groups.items(), key=lambda v: v[1]):
            print("{}\t{}".format(k, v))

    print("TOTAL\t{}".format(total))


@click.command(short_help='Groupby and run script')
@click.argument('input', type=click.Path(exists=True), nargs=1)
@click.option('--script', '-s', type=click.STRING)
@click.option('--where', '-w', multiple=True, type=click.STRING)
@click.option('--groupby', '-g', type=click.STRING)
@click.option('--cores', help='Maximum processes to run in parallel.', type=click.INT, default=os.cpu_count())
@click.option('--quite', '-q', help="Don't show the progress, only the total count.", is_flag=True)
def groupby(input, script, where, groupby, cores, quite):

    selection = list(selector.groupby(input, by=groupby, where=__parse_where(where)))

    with Pool(cores) as pool:
        task = functools.partial(tasks.script, script=script)
        map_method = pool.imap_unordered if len(selection) > 1 else map

        for group_key, group_result in tqdm(
                map_method(task, selection),
                total=len(selection),
                desc="Computing groups".rjust(40),
                disable=(len(selection) < 2 or quite)
        ):
            for r in group_result:
                print("{}\t{}".format(group_key, r))


cli.add_command(preprocess)
cli.add_command(cat)
cli.add_command(count)
cli.add_command(groupby)

if __name__ == "__main__":
    cli()




import gzip
import lzma


def __base_parser(lines, count=False):
    for l, line in enumerate(lines, start=1):
        # Skip empty lines
        if len(line) == 0:
            continue

        # Skip comments
        if line.startswith('#') and not line.startswith('#CHROM'):
            continue

        # Parse columns
        if not count:
            line = [v.strip() for v in line.split('\t')]

        yield l, line


def __open_file(file, mode="rt"):
    close = False
    fd = file

    # A file path
    if isinstance(file, str):

        # Detect open method
        open_method = open
        if file.endswith('gz') or file.endswith('bgz'):
            open_method = gzip.open
        elif file.endswith('xz'):
            open_method = lzma.open

        fd = open_method(file, mode)
        close = True

    return fd, close


def __count_lines(file):
    fd, close = __open_file(file)
    result = sum(1 for _ in __base_parser(fd, count=True))
    if close:
        fd.close()
    return result


from bgparsers import readers
from subprocess import Popen, PIPE

HEADERS = ["SAMPLE", "DONOR", "CHROMOSOME", "POSITION", "REF", "ALT", "STRAND", "ALT_TYPE"]


def __skip(row, where):
    # TODO Implement where using annotations

    if where is None:
        return False

    for k, v in where.items():
        if str(row.get(k, None)) != v:
            return True
    return False


def run(selection, script=None):
    group_key, group_values = selection

    process = Popen(script, shell=True, stdin=PIPE, stdout=PIPE, env={"GROUP_KEY": group_key})
    try:
        for r in readers.variants(group_values):
            process.stdin.write("{}\n".format("\t".join([str(r.get(h, "")) for h in HEADERS])).encode())
            process.stdin.flush()
        process.stdin.close()
    except BrokenPipeError:
        pass

    output = []
    try:
        while True:
            out = process.stdout.readline().decode().strip()
            if out == "":
                break
            output.append(out)
        process.stdout.close()
    except BrokenPipeError:
        pass

    return group_key, output





from bgparsers import readers


def __skip(row, where):
    if where is None:
        return False

    for k, v in where.items():
        if str(row.get(k, None)) != v:
            return True
    return False


def run(selection, columns=None, where=None):

    try:
        headers = ["SAMPLE", "DONOR", "CHROMOSOME", "POSITION", "REF", "ALT", "STRAND", "ALT_TYPE"] + list(columns)
        print("\t".join(headers))
        for input_file, input_annotations in selection:
            for r in readers.variants(input_file, annotations=input_annotations):

                # TODO Implement where using annotations
                if __skip(r, where):
                    continue

                print("\t".join([str(r.get(h, "")) for h in headers]))
    except BrokenPipeError:
        pass




import csv
import gzip
import os
import glob
import re

from collections import defaultdict
from fnmatch import fnmatch
from os.path import join, isfile, dirname
from configobj import ConfigObj

from bgparsers import readers

ANNOTATION_FILE_EXTENSION = "bginfo"
EXCLUDE_PREFIX = "__exclude_"


def __null_value(value):
    return None if value == "" else value


def __read_annotations(annotation_file):

    info = ConfigObj(annotation_file)

    extensions = info['pattern'] if isinstance(info['pattern'], list) else [info['pattern'].strip()]
    recursive = info.get('recursive', 'false').lower() in ["true", "yes", "y", "1", "ok"]
    annotations = info.get('annotations', {})
    excludes = info.get('excludes', {})

    for k in annotations.keys():
        v = annotations[k]
        if isinstance(v, str) and v.startswith('(') and v.endswith(')'):
            annotate = eval(v)
            if annotate[0] == 'mapping':
                mapping_files = join(dirname(annotation_file), annotate[2])
                values = {}
                for mapping_file in glob.glob(mapping_files):
                    open_method = gzip.open if mapping_file.endswith('gz') else open
                    with open_method(mapping_file, "rt") as fd:
                        for r in csv.DictReader(fd, delimiter='\t'):
                            a, b = __null_value(r[annotate[3]]), __null_value(r[annotate[4]])
                            if a is not None and b is not None:
                                if len(annotate) > 5:
                                    b = annotate[5](b)
                                values[a] = b
                annotations[k] = ('mapping', annotate[1], values)
            if annotate[0] == 'filename':
                annotations[k] = ('filename', re.compile(annotate[1]), (lambda v: v) if len(annotate) == 2 else annotate[2])
            if annotate[0] == 'dirname':
                annotations[k] = ('dirname', re.compile(annotate[1]), (lambda v: v) if len(annotate) == 2 else annotate[2])
            if annotate[0] == 'internal':
                annotations[k] = ('internal', annotate[1])
        else:
            annotations[k] = __null_value(v)

    for k, v in excludes.items():
        v = v if isinstance(v, list) or isinstance(v, set) else {v}
        annotations["{}{}".format(EXCLUDE_PREFIX, k)] = set([__null_value(i) for i in v])

    return recursive, {e: annotations for e in extensions}


def __merge_annotations(a, b):
    """

    :param a: The first annotations dictionary
    :param b: The second annotations dictionary. This annotations have preference and will override A annotations if there is a conflict
    :return: The union of A and B annotation dictionaries
    """

    # Clone A
    c = {k: dict(v) for k, v in a.items()}

    # Update or add B entries
    for k, v in b.items():
        if k in c:

            # Update the annotations
            for kv, vv in v.items():
                if kv not in c[k]:
                    c[k][kv] = vv
                else:
                    if isinstance(c[k][kv], list):
                        # If it's a list concat them instead of override it
                        vv_list = vv if isinstance(vv, list) else [vv]
                        c[k][kv] = list(c[k][kv]) + vv_list
                    else:
                        # Override the value
                        c[k][kv] = vv

        else:
            c[k] = dict(v)

    return c


def __where_cmp(data_value, where_value):
    if isinstance(where_value, list) or isinstance(where_value, set):
        return data_value in where_value
    else:
        return data_value != where_value


def __where_match(annotations, where=None):

    if where is None:
        return True

    for k, v in where.items():
        if k in annotations:
            value = annotations[k]
            if isinstance(value, list):
                if v not in value:
                    return False
            elif isinstance(value, tuple) and value[0] == 'mapping':
                if isinstance(v, list) or isinstance(v, set):
                    filtered_mapping = {mk: mv for mk, mv in value[2].items() if mv in v}
                    annotations[k] = ('filtering', value[1], filtered_mapping)
                    if len(filtered_mapping) == 0:
                        return False
                else:
                    where_valid_values = set([nk for nk, nv in value[2].items() if nv == v])
                    annotations[k] = (value[1], v, where_valid_values)
                    if len(where_valid_values) == 0:
                        return False
            else:
                if not __where_cmp(value, v):
                    return False
        else:
            return False

    return True


def __process_annotations(annotations, basepath, filename):
    for k in annotations:
        value = annotations[k]
        if isinstance(value, tuple):
            if value[0] in ['filename', 'dirname']:
                name = filename if value[0] == 'filename' else os.path.basename(basepath)
                m = value[1].match(name)
                annotations[k] = None
                if m is not None:
                    g = m.groups()
                    annotations[k] = value[2](*g).format(**annotations)

    return annotations


def find(basepath, where=None, annotations=None):

    global_annotations = {} if annotations is None else annotations

    # Annotations that only apply to the current folder level
    local_annotations = {}

    # Update annotations
    for annotation_file in glob.iglob(join(basepath, "*.{}".format(ANNOTATION_FILE_EXTENSION))):
        recursive, ann = __read_annotations(annotation_file)

        # Update global annotations
        if recursive:
            global_annotations = __merge_annotations(global_annotations, ann)
        else:
            local_annotations = __merge_annotations(local_annotations, ann)

    # Merge local and global annotations, but local have preference over global
    local_annotations = __merge_annotations(global_annotations, local_annotations)

    # Find files
    for file in os.listdir(basepath):
        file_path = join(basepath, file)

        if isfile(file_path):
            file_annotations = {}
            pattern_matches = 0
            for pattern, annotations in local_annotations.items():
                if fnmatch(file_path, pattern):
                    pattern_matches += 1
                    file_annotations.update(annotations)

            # Process filename annotations
            file_annotations = __process_annotations(file_annotations, basepath, file)

            if pattern_matches > 0 and __where_match(file_annotations, where=where):
                yield file_path, file_annotations

        else:
            # Search subfolders
            for f, a in find(file_path, where=where, annotations=global_annotations):
                yield f, a


def __get_unique_values(file, key):
    values = set()
    for r in readers.variants(file):
        values.add(r[key])
    return values


def get_samples_ids(file):
    return __get_unique_values(file, 'SAMPLE')


def get_donor_ids(file):
    return __get_unique_values(file, 'DONOR')


def groupby(basepath, by=None, where=None, no_group_by=True):
    assert by != None, "You must provide a 'by' annotation"

    results = defaultdict(list)

    for f, a in find(basepath, where=where):
        by_value = a.get(by, None)

        if no_group_by:
            if by_value is None or isinstance(by_value, tuple):
                a_clone = dict(a)
                a_clone[by] = ('not', by_value)
                results[None].append((f, a_clone))

        if by_value is None:
            continue

        if isinstance(by_value, tuple):

            if by_value[0] == 'internal':
                values = by_value[1]
                if not isinstance(values, set):
                    if values == 'SAMPLE':
                        values = get_samples_ids(f)
                        a[by] = ('internal', values)
                    elif values == 'DONOR':
                        values = get_donor_ids(f)
                        a[by] = ('internal', values)
                    else:
                        raise NotImplementedError("'{}' internal annotation not implemented".format(values))

                for s in values:
                    a_clone = dict(a)
                    a_clone[by] = (by, s, {s})
                    results[s].append((f, a_clone))

            elif isinstance(by_value[2], dict):
                inv_dict = {}
                for k, v in by_value[2].items():
                    inv_dict[v] = inv_dict.get(v, set())
                    inv_dict[v].add(k)

                for k, values in inv_dict.items():
                    a_clone = dict(a)
                    a_clone[by] = (by_value[1], k, values)
                    results[k].append((f, a_clone))
            else:
                results[by_value[1]].append((f, a))

        else:
            results[by_value].append((f, a))

    for key, group in results.items():
        yield key, group




















from bgparsers import readers


def __skip(row, where):
    # TODO Implement where using annotations

    if where is None:
        return False

    for k, v in where.items():
        if str(row.get(k, None)) != v:
            return True
    return False


def run(selection, where=None, groupby=None):

    i = 0
    input_file, input_annotations = selection

    if groupby is None:
        for r in readers.variants(input_file, annotations=input_annotations):

            if __skip(r, where):
                continue

            i += 1
        return i, None
    else:
        groups = {}
        for r in readers.variants(input_file, annotations=input_annotations):

            if __skip(r, where):
                continue

            val = r[groupby]
            val_count = groups.get(val, 0)
            groups[val] = val_count + 1
            i += 1

        return i, groups





import lzma
import pickle

import logging
import numpy as np

from os import unlink, makedirs
from os.path import join, exists, dirname, basename
from bgparsers import readers


SEGMENT_SIZE_MIN = 10000
SEGMENT_SIZE = 2000000


def unify_column(values):
    val = values[0]
    for v in values:
        if v != val:
            return values
    return val


def minify_column(values):
    i = 0
    meta = {}

    result = []
    for v in values:
        if v not in meta:
            meta[v] = i
            i += 1
        result.append(meta[v])

    meta_list = [None] * i
    for k, v in meta.items():
        meta_list[v] = k

    # NumPY data types
    # uint8     Unsigned integer(0 to 255)
    # uint16    Unsigned integer(0 to 65535)
    # uint32    Unsigned integer(0 to 4294967295)
    # uint64    Unsigned integer(0 to 18446744073709551615)

    size = len(meta_list)
    if size <= 256:
        dtype = np.uint8
    elif size <= 65536:
        dtype = np.uint16
    else:
        dtype = np.uint32

    return meta_list, np.array(result, dtype=dtype)


def store_chunk(chunk, output_path, results, length):

    file_path = join(output_path, "c{:06d}.bgvars.xz".format(chunk))

    for k in results:
        results[k] = unify_column(results[k])

    values = {}
    for k in results:
        if isinstance(results[k], list):
            values[k], results[k] = minify_column(results[k])

    out_data = {
        'length': length,
        'data': values,
        'vectors': results
    }

    if len(results) > 0:
        try:
            with lzma.open(file_path, 'wb') as fd:
                pickle.dump(out_data, fd)
        except Exception as e:
            if exists(file_path):
                unlink(file_path)
            raise RuntimeError(e)


def run(input_selection, output_folder=None, force=False, exist_ok=False):
    input_file, input_annotations = input_selection

    if output_folder is None:
        output_folder = join(dirname(input_file), "bgvariants", "preprocess", basename(input_file))

    if force or not exists(output_folder):
        results = {}
        c = 0
        chunk = 0

        seg_size = None
        makedirs(output_folder, exist_ok=True)
        for r in readers.variants(input_file, annotations=input_annotations, extra=True):

            if seg_size is None:
                seg_size = max(SEGMENT_SIZE / (len(r) / 8), SEGMENT_SIZE_MIN)

            for k in r.keys():
                if k not in results:
                    results[k] = [None]*c
            for k in results.keys():
                results[k].append(r.get(k, None))
            c+=1

            if c >= seg_size:
                store_chunk(chunk, output_folder, results, c)
                chunk += 1
                c = 0
                results = {}

        if c > 0:
            store_chunk(chunk, output_folder, results, c)
    else:
        if exist_ok:
            logging.warning("Preprocessed folder '{}' already exists".format(basename(output_folder)))
        else:
            raise FileExistsError("Preprocess folder '{}' already exists".format(output_folder))




import types
import pickle
import logging
import numpy as np

logger = logging.getLogger("bgparsers")

from os.path import exists, join, dirname, basename
from bgparsers.readers.common import __open_file, __count_lines, __base_parser

EXCLUDE_PREFIX = "__exclude_"

COMMON_HEADERS = ['CHROMOSOME', 'POSITION', 'STRAND', 'REF', 'ALT', 'ALT_TYPE', 'SAMPLE', 'DONOR']
TSV_HEADERS = [

    # Chromosome
    ({'chromosome', 'chr', 'chrom', 'chromosome_name', '#chrom'}, 'CHROMOSOME',
     lambda c: c.upper().replace('CHR', '').replace('23', 'X').replace('24', 'Y')),

    # Start position
    ({'position', 'start', 'start_position', 'pos', 'chromosome_start'}, 'POSITION', int),

    # Strand
    ({'strand', 'chromosome_strand'}, 'STRAND',
     lambda s: '-' if s in ['-', '0', '-1'] else '+' if s in ['+', '1', '+1'] else None),

    # Reference
    ({'ref', 'reference_allele', 'reference', 'reference_genome_allele'}, 'REF', str),

    # Alternate
    ({'alt', 'tumor_seq_allele2', 'variant', 'alternative', 'mutated_to_allele'}, 'ALT', str),

    # Sample ID
    ({'sample', 'tumor_sample_barcode'}, 'SAMPLE', str),

    # TCGA only: donor deduced from the sample id
    ({'tumor_sample_barcode'}, 'DONOR', lambda s: "-".join(s.split("-")[:3]) if s.startswith("TCGA") else None),

    # Donor ID
    ({'donor_id'}, 'DONOR', str)
]


def __known_headers(header):
    h_lower = header.lower()
    return [(field, method) for valid_headers, field, method in TSV_HEADERS if h_lower in valid_headers]


def tsv_header(line, extra=False, required=None):
    header = []
    extra_header = []
    for i, h in enumerate(line):
        known = __known_headers(h)
        if len(known) > 0:
            for field, method in known:
                header.append((field, method, i))
        else:
            extra_header.append((h, str, i))

    inferred_header = len(header) == 0
    if inferred_header:
        # TODO Infer the header from the values of the first line
        raise NotImplementedError("We need a header")

    if extra:

        # TODO Check header collision
        if isinstance(extra, list):
            header += [h for h in extra_header if h[0] in extra]
        else:
            header += extra_header

        if required is not None and not (set(required) <= set([h[0] for h in header])):
            raise SyntaxError('Missing fields in file header. Required fields: {}'.format(required))

    return header, inferred_header


def tsv_parser(lines, header=None, close=False, extra=False, required=None, name=""):
    for l, line in __base_parser(lines):

        # Initialize header
        if header is None:
            header, inferred_header = tsv_header(line, extra=extra, required=required)
            if not inferred_header:
                continue

        try:
            row = {h[0]: h[1](line[h[2]]) for h in header}
        except (ValueError, IndexError) as e:
            logger.warning("Error parsing line %d %s (%s %s %s)", l, name, e, line, header)
            continue

        yield row

    if close:
        lines.close()


def chunks_parser(data_folder, extra=False):
    chunk_count = 0
    chunk_file = join(data_folder, "c{:06d}.bgvars.xz".format(chunk_count))
    while exists(chunk_file):
        for r in pickle_parser(chunk_file, extra=extra):
            yield r
        chunk_count += 1
        chunk_file = join(data_folder, "c{:06d}.bgvars.xz".format(chunk_count))


def pickle_parser(file, extra=False):
    fd, close = __open_file(file, "rb")
    values = pickle.load(fd)
    if close:
        fd.close()

    if isinstance(values, dict):
        size = values['length']
        data = values['data']
        vectors = values['vectors']
        base = dict([(k, vectors[k]) for k in vectors.keys() if not isinstance(vectors[k], np.ndarray)])
        keys = [k for k in vectors.keys() if isinstance(vectors[k], np.ndarray)]

        if isinstance(extra, list) or not extra:
            fields = set(COMMON_HEADERS + extra) if isinstance(extra, list) else set(COMMON_HEADERS)
            keys = [k for k in keys if k in fields]

        for i in range(size):
            r = base.copy()
            for k in keys:
                r[k] = data[k][vectors[k][i]]
            yield r

    else:
        if isinstance(extra, list) or not extra:
            fields = COMMON_HEADERS + extra if isinstance(extra, list) else COMMON_HEADERS
            for r in values:
                yield {k: r[k] for k in fields}
        else:
            for r in values:
                yield r


def __is_pickle(file):
    return file.endswith(".bgvars.gz") or file.endswith(".bgvars.xz") or file.endswith(".bgvars")


def __parser_generator(file, extra=False, required=None):
    pickle_version = __preprocess_file(file)
    if exists(pickle_version):
        return chunks_parser(dirname(pickle_version), extra=extra)

    if __is_pickle(file):
        return pickle_parser(file, extra=extra)

    fd, close = __open_file(file, "rt")
    return tsv_parser(fd, close=close, extra=extra, required=required, name=file)


def __preprocess_file(file):
    return join(dirname(file), "bgvariants", "preprocess", basename(file), "c000000.bgvars.xz")


def variants_count(file):
    return __count_lines(file)


def variants(file, annotations=None, extra=False, required=None):
    try:
        if isinstance(file, types.GeneratorType) or isinstance(file, list):
            # Allow to concatenate a reader and a selector. Ex: readers.variants( selector.find( ...
            for f, a in file:
                for row in variants(f, annotations=a, extra=extra, required=required):
                    yield row
        else:

            preprocess = exists(__preprocess_file(file)) or __is_pickle(file)
            if annotations is not None and preprocess:
                annotations = {k: v for k, v in annotations.items() if
                               isinstance(v, tuple) and v[0] not in ['mapping', 'internal']}

            if annotations is not None:
                internal_annotations = []
                for k, v in annotations.items():
                    if isinstance(v, tuple) and v[0] == 'internal':
                        if isinstance(v[1], tuple):
                            internal_annotations += v[1][1]
                        else:
                            internal_annotations.append(v[1])

                if len(internal_annotations) > 0 and extra != True:
                    if isinstance(extra, list):
                        extra += [h for h in internal_annotations if h not in extra]
                    else:
                        extra = internal_annotations

            parser = __parser_generator(file, extra=extra, required=required)
            if annotations is None or len(annotations) == 0:
                for row in parser:

                    # Add postprocessing annotations
                    if not preprocess:
                        row = __postprocess_row(row)

                    if row is None:
                        continue

                    yield row
            else:
                for row in parser:

                    # Add row annotations and filter non-valid rows
                    row = __annotate_row(row, annotations)

                    if row is None:
                        continue

                    # Add postprocessing annotations
                    if not preprocess:
                        row = __postprocess_row(row)

                    yield row
    except NotImplementedError as e:
        logger.error("{}. {}".format(file, e))
        raise e


def prefix_length(ref, alt):
    i = 0
    while i < len(ref) and i < len(alt) and ref[i] == alt[i]:
        i += 1
    return i


def suffix_length(ref, alt):
    i = len(ref) - 1
    j = len(alt) - 1
    while i >= 0 and j >= 0 and ref[i] == alt[j]:
        i -= 1
        j -= 1
    return len(ref) - i - 1


def indel_postprocess(start, ref, alt):
    """
    Removes the bases that are repeated in both ref and alt and are therefore NOT variants
    :param start:
    :param ref:
    :param alt:
    :return:
    """
    prefix_len = prefix_length(ref, alt)
    ins_correction = 1 if len(ref) < len(alt) else 0
    start += max(0, prefix_len - ins_correction)
    alt = alt[prefix_len:]
    ref = ref[prefix_len:]

    suffix_len = suffix_length(ref, alt)
    if suffix_len > 0:
        alt = alt[:-suffix_len]
        ref = ref[:-suffix_len]

    ref = '-' if ref == '' else ref
    alt = '-' if alt == '' else alt

    return start, ref, alt


def __postprocess_row(row):
    if "ALT_TYPE" not in row:
        l_ref = len(row['REF'])
        l_alt = len(row['ALT'])

        alteration_type = None
        if l_alt != l_ref:
            alteration_type = "indel"
        else:
            if l_alt > 1:
                alteration_type = "mnp"
            else:
                if '-' in row['REF'] or '-' in row['ALT']:
                    alteration_type = "indel"
                else:
                    alteration_type = "snp"
        row['ALT_TYPE'] = alteration_type

    if "STRAND" not in row:
        row['STRAND'] = "+"

    # Use VEP like insertions and deletions formatting
    if row['ALT_TYPE'] == "indel":
        row['POSITION'], row['REF'], row['ALT'] = indel_postprocess(row['POSITION'], row['REF'], row['ALT'])

    return row


def __annotate_row(row, annotations):
    excludes = {}

    sorted_annotations = list(sorted(annotations.items(), key=lambda v: 1 if not isinstance(v[1], tuple) else 2 if v[1][0] == 'internal' else 3 if v[1][0] == 'mapping' else 4))

    for k, v in sorted_annotations:

        # Skip exclusion
        if k.startswith(EXCLUDE_PREFIX):
            excludes[k.replace(EXCLUDE_PREFIX, "")] = v
            continue

        if isinstance(v, tuple):

            if v[0] == 'mapping':
                row[k] = __annotate_value(row, v)
            elif v[0] == 'filtering':
                row[k] = __annotate_value(row, v)
                if row[k] is None:
                    return None
            elif v[0] == 'not':
                v_not = v[1]
                if isinstance(v_not, tuple):
                    if v_not[0] == 'mapping':
                        v_not = __annotate_value(row, v)
                if v_not is not None:
                    return None
            elif v[0] == 'internal':
                if isinstance(v[1], tuple):
                    try:
                        row[k] = v[1][0].format(**row)
                    except Exception:
                        row[k] = None
                else:
                    row[k] = row.get(v[1], None)
                continue
            else:
                # Filter non annotated rows
                if row[v[0]] not in v[2]:
                    return None
                row[k] = v[1]
        else:
            row[k] = v

    # Process exclusions after the row has been annotated
    for k, v in excludes.items():
        if row[k] in v:
            return None

    return row


def __annotate_value(row, annotation):
    map_key = row[annotation[1]]
    map_value = annotation[2].get(map_key, None)
    return map_value




def dequote(s):
    """
    If a string has single or double quotes around it, remove them.
    Make sure the pair of quotes match.
    If a matching pair of quotes is not found, return the string unchanged.
    """
    if (s[0] == s[-1]) and s.startswith(("'", '"')):
        return s[1:-1]
    return s


def chunkizator(iterable, size=1000):
    """
    Creates chunks from an iterable

    Args:
        iterable:
        size (int): elements in the chunk

    Returns:
        list. Chunk

    """
    s = 0
    chunk = []
    for i in iterable:
        if s == size:
            yield chunk
            chunk = []
            s = 0
        chunk.append(i)
        s += 1
    yield chunk


from bgparsers import __version__
from setuptools import setup, find_packages

setup(
    name='bgparsers',
    version=__version__,
    packages=find_packages(),
    author='BBGLab (Barcelona Biomedical Genomics Lab)',
    author_email='bbglab@irbbarcelona.org',
    description="Bgconfig library to manage configuration files.",
    license="Apache License 2",
    keywords="",
    url="https://bitbucket.org/bgframework/bgconfig",
    download_url="https://bitbucket.org/bgframework/bgconfig/get/"+__version__+".tar.gz",
    install_requires=[
        "tqdm",
        "bgconfig",
        "itab",
        "click",
        "numpy",
        "intervaltree"
    ],
    entry_points={
            'console_scripts': [
                'bgvariants = bgparsers.commands.bgvariants:cli',
            ]
        }
)